lib LibC
  TCP_NODELAY      =    1
  TCP_TXTLS_ENABLE =   39
  TCP_RXTLS_ENABLE =   41
  TCP_KEEPIDLE     =  256
  TCP_KEEPINTVL    =  512
  TCP_KEEPCNT      = 1024

  TLS_SET_RECORD_TYPE = 1
  TLS_GET_RECORD      = 2
end
